from typing import Any, Optional, Dict, Generator, AsyncGenerator
from copy import deepcopy
from llama_index.core.async_utils import run_async_tasks
from llama_index.core.base.base_query_engine import BaseQueryEngine
from llama_index.core.base.response.schema import (
    RESPONSE_TYPE,
    Response,
    AsyncStreamingResponse,
    StreamingResponse,
)
from llama_index.core.callbacks.base import CallbackManager
from llama_index.core.base.llms.types import ChatMessage, MessageRole
from llama_index.core.prompts.base import ChatPromptTemplate, BasePromptTemplate
from llama_index.core.prompts.mixin import PromptDictType, PromptMixinType
from llama_index.core.schema import NodeWithScore, QueryBundle, TextNode
from llama_index.indices.managed.postgresml.retriever import PostgresMLRetriever


PROMPTS = {
    "text_qa_template": ChatPromptTemplate(
        message_templates=[
            ChatMessage(content="You are a helpful chatbot", role=MessageRole.SYSTEM),
            ChatMessage(
                content="""Context information is below.
---------------------
{context_str}
---------------------
Given the context information and not prior knowledge,
Query: {query_str}
Answer: """,
                role=MessageRole.USER,
            ),
        ]
    )
}


class AsyncJsonGenerator(Generator, AsyncGenerator):
    def __init__(self, rag_stream_results) -> None:
        self.rag_stream_results = rag_stream_results
        self.rag_stream = None

    def asend(self):
        raise Exception("asend is not implemented")

    def send(self):
        raise Exception("send is not implemented")

    def athrow(self):
        raise Exception("athrow is not implemented")

    def throw(self):
        raise Exception("throw is not implemented")

    def __iter__(self) -> "AsyncJsonGenerator":
        return self

    def __aiter__(self) -> "AsyncJsonGenerator":
        return self

    def __next__(self) -> str:
        try:
            return run_async_tasks([self.__anext__()])[0]
        except StopAsyncIteration:
            raise StopIteration

    async def __anext__(self) -> str:
        if not self.rag_stream:
            self.rag_stream = self.rag_stream_results.stream()
        result = await self.rag_stream.__anext__()
        if len(result) > 0:
            return result[0]
        else:
            return ""


class PostgresMLQueryEngine(BaseQueryEngine):
    """Retriever query engine for PostgresML."""

    def __init__(
        self,
        retriever: PostgresMLRetriever,
        streaming: Optional[bool] = False,
        callback_manager: Optional[CallbackManager] = None,
        pgml_query: Optional[Dict[str, Any]] = None,
        vector_search_limit: Optional[int] = 4,
        vector_search_rerank: Optional[Dict[str, Any]] = None,
        vector_search_document: Optional[Dict[str, Any]] = {"keys": ["id", "metadata"]},
        model: Optional[str] = "meta-llama/Meta-Llama-3-8B-Instruct",
        model_parameters: Optional[Dict[str, Any]] = {"max_tokens": 2048},
        **kwargs,
    ) -> None:
        self._retriever = retriever
        self._streaming = streaming
        self._prompts = deepcopy(PROMPTS)
        self._pgml_query = pgml_query
        self._vector_search_limit = vector_search_limit
        self._vector_search_rerank = vector_search_rerank
        self._vector_search_document = vector_search_document
        self._model = model
        self._model_parameters = model_parameters
        super().__init__(callback_manager=callback_manager)

    @classmethod
    def from_args(
        cls,
        retriever: PostgresMLRetriever,
        **kwargs: Any,
    ) -> "PostgresMLQueryEngine":
        """Initialize a PostgresMLQueryEngine object.".

        Args:
            retriever (PostgresMLRetriever): A PostgresML retriever object.
        """
        return cls(retriever=retriever, **kwargs)

    def with_retriever(self, retriever: PostgresMLRetriever) -> "PostgresMLQueryEngine":
        return PostgresMLQueryEngine(
            retriever=retriever,
        )

    def _query(self, query_bundle: QueryBundle) -> RESPONSE_TYPE:
        """Answer a query."""
        if self._streaming:
            async_token_gen = run_async_tasks([self._do_query(query_bundle)])[0]
            return StreamingResponse(response_gen=async_token_gen.response_gen)
        else:
            return run_async_tasks([self._do_query(query_bundle)])[0]

    async def _aquery(self, query_bundle: QueryBundle, **kwargs) -> RESPONSE_TYPE:
        """Answer an async query."""
        return await self._do_query(query_bundle, **kwargs)

    async def _do_query(
        self, query_bundle: Optional[QueryBundle] = None
    ) -> RESPONSE_TYPE:
        query = self._pgml_query
        if not query:
            if not query_bundle:
                raise Exception(
                    "Must provide either query or query_bundle to query and aquery"
                )

            # {CONTEXT} gets replaced with the correct context in the SQL query generated by the pgml SDK
            messages = self._prompts["text_qa_template"].format_messages(
                context_str="{CONTEXT}", query_str=query_bundle.query_str
            )
            messages = [
                {"role": m["role"].value, "content": m["content"]}
                for m in [m.dict() for m in messages]
            ]

            model_parameters = deepcopy(self._model_parameters)
            model_parameters["model"] = self._model
            model_parameters["messages"] = messages

            if self._vector_search_rerank is not None:
                self._vector_search_rerank = self._vector_search_rerank | {
                    "query": query_bundle.query_str
                }

            query = {
                "CONTEXT": {
                    "vector_search": {
                        "query": {
                            "fields": {
                                "content": {
                                    "query": query_bundle.query_str,
                                    "parameters": {"prompt": "query: "},
                                },
                            },
                        },
                        "document": self._vector_search_document,
                        "limit": self._vector_search_limit,
                        "rerank": self._vector_search_rerank,
                    },
                    "aggregate": {"join": "\n"},
                },
                "chat": model_parameters,
            }

        if self._streaming:
            # The pgml SDK does not currently return sources for streaming
            results = await self._retriever._index.collection.rag_stream(
                query,
                self._retriever._index.pipeline,
            )
            return AsyncStreamingResponse(response_gen=AsyncJsonGenerator(results))
        else:
            results = await self._retriever._index.collection.rag(
                query,
                self._retriever._index.pipeline,
            )
            source_nodes = [
                NodeWithScore(
                    node=TextNode(
                        id_=r["document"]["id"],
                        text=r["chunk"],
                        metadata=r["document"]["metadata"],
                    ),
                    score=r["score"],
                )
                if self._vector_search_rerank is None
                else NodeWithScore(
                    node=TextNode(
                        id_=r["document"]["id"],
                        text=r["chunk"],
                        metadata=r["document"]["metadata"],
                    ),
                    score=r["rerank_score"],
                )
                for r in results["sources"]["CONTEXT"]
            ]
            return Response(response=results["rag"][0], source_nodes=source_nodes)

    def _get_prompt_modules(self) -> PromptMixinType:
        """Get prompt modules."""
        return {}

    def _get_prompts(self) -> Dict[str, BasePromptTemplate]:
        """Get prompts."""
        return self._prompts

    def _update_prompts(self, prompts_dict: PromptDictType):
        """Update prompts."""
        for key in prompts_dict:
            self._prompts[key] = prompts_dict[key]



"""PostgresML index.
An index that is built on top of PostgresML.
"""

import logging
from typing import Any, List, Optional, Dict

from llama_index.core.async_utils import run_async_tasks
from llama_index.core.base.base_retriever import BaseRetriever
from llama_index.core.callbacks.base import CallbackManager
from llama_index.core.schema import NodeWithScore, QueryBundle, TextNode
from llama_index.indices.managed.postgresml.base import PostgresMLIndex

_logger = logging.getLogger(__name__)


class PostgresMLRetriever(BaseRetriever):
    """PostgresML Retriever.

    Args:
        index (PostgresMLIndex): the PostgresML Index
    """

    def __init__(
        self,
        index: PostgresMLIndex,
        callback_manager: Optional[CallbackManager] = None,
        pgml_query: Optional[Dict[str, Any]] = None,
        limit: Optional[int] = 5,
        rerank: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> None:
        """Initialize params."""
        self._index = index
        self._pgml_query = pgml_query
        self._limit = limit
        self._rerank = rerank
        super().__init__(callback_manager)

    def _retrieve(
        self,
        query_bundle: Optional[QueryBundle] = None,
        **kwargs: Any,
    ) -> List[NodeWithScore]:
        return run_async_tasks([self._aretrieve(query_bundle, **kwargs)])[0]

    async def _aretrieve(
        self,
        query_bundle: Optional[QueryBundle] = None,
    ) -> List[NodeWithScore]:
        async def do_vector_search():
            if self._pgml_query:
                return await self._index.collection.vector_search(
                    self._pgml_query,
                    self._index.pipeline,
                )
            else:
                if not query_bundle:
                    raise Exception(
                        "Must provide either query or query_bundle to retrieve and aretrieve"
                    )
                if self._rerank is not None:
                    self._rerank = self._rerank | {"query": query_bundle.query_str}
                return await self._index.collection.vector_search(
                    {
                        "query": {
                            "fields": {
                                "content": {
                                    "query": query_bundle.query_str,
                                    "parameters": {"prompt": "query: "},
                                }
                            }
                        },
                        "rerank": self._rerank,
                        "limit": self._limit,
                    },
                    self._index.pipeline,
                )

        results = await do_vector_search()
        return [
            NodeWithScore(
                node=TextNode(
                    id_=r["document"]["id"],
                    text=r["chunk"],
                    metadata=r["document"]["metadata"],
                ),
                score=r["score"],
            )
            if self._rerank is None
            else NodeWithScore(
                node=TextNode(
                    id_=r["document"]["id"],
                    text=r["chunk"],
                    metadata=r["document"]["metadata"],
                ),
                score=r["rerank_score"],
            )
            for r in results
        ]



"""Managed index.

A managed Index - where the index is accessible via some API that
interfaces a managed service.

"""

import logging
from llama_index.core.async_utils import run_async_tasks
from typing import Any, Dict, Optional, Sequence, Type

from llama_index.core.base.base_query_engine import BaseQueryEngine
from llama_index.core.base.base_retriever import BaseRetriever

from llama_index.core.data_structs.data_structs import IndexDict, IndexStructType
from llama_index.core.indices.managed.base import BaseManagedIndex, IndexType
from llama_index.core.schema import (
    BaseNode,
    Document,
    TextNode,
)
from pgml import Collection, Pipeline, init_logger

init_logger()

_logger = logging.getLogger(__name__)


class PostgresMLIndexStruct(IndexDict):
    """PostgresML Index Struct."""

    @classmethod
    def get_type(cls) -> IndexStructType:
        """Get index struct type."""
        # return IndexStructType.POSTGRESML
        return "POSTGRESML"


class PostgresMLIndex(BaseManagedIndex):
    """PostgresML Index.

    The PostgresML index implements a managed index that uses PostgresML as the backend.
    PostgresML performs a lot of the functions in traditional indexes in the backend:
    - breaks down a document into chunks (nodes)
    - Creates the embedding for each chunk (node)
    - Performs the search for the top k most similar nodes to a query
    - Optionally can perform text-generation or chat completion
    """

    def __init__(
        self,
        collection_name: str,
        pipeline_name: Optional[str] = None,
        pipeline_schema: Optional[Dict[str, Any]] = None,
        pgml_database_url: Optional[str] = None,
        show_progress: bool = True,
        upsert_parallel_batches: int = 1,
        nodes: Optional[Sequence[BaseNode]] = None,
        **kwargs: Any,
    ) -> None:
        """Initialize the PostgresML SDK."""
        self.show_progress = show_progress
        self.upsert_parallel_batches = upsert_parallel_batches

        index_struct = PostgresMLIndexStruct(
            index_id=collection_name,
            summary="PostgresML Index",
        )

        super().__init__(
            show_progress=show_progress,
            index_struct=index_struct,
            **kwargs,
        )

        # Create our Collection and Pipeline
        self.collection = Collection(collection_name, pgml_database_url)
        if pipeline_name is None:
            pipeline_name = "v1"
        if pipeline_schema is None:
            pipeline_schema = {
                "content": {
                    "splitter": {
                        "model": "recursive_character",
                        "parameters": {"chunk_size": 1500},
                    },
                    "semantic_search": {
                        "model": "intfloat/e5-small-v2",
                        "parameters": {"prompt": "passage: "},
                    },
                }
            }
        self.pipeline = Pipeline(pipeline_name, pipeline_schema)

        # We must wrap self.collection.add_pipeline() with this async function
        # This is a limitation of the pyo3 async implementation
        async def add_pipeline():
            await self.collection.add_pipeline(self.pipeline)

        run_async_tasks([add_pipeline()])

        if nodes:
            self._insert(nodes)

    def _insert(
        self,
        nodes: Sequence[BaseNode],
        **insert_kwargs: Any,
    ) -> None:
        """Insert a set of documents (each a node)."""
        documents = [
            {
                "id": node.node_id,
                "content": node.get_content(),
                "metadata": node.metadata,
            }
            for node in nodes
        ]

        args = {"parallel_batches": self.upsert_parallel_batches, **insert_kwargs}

        # We must wrap self.collection.upsert_documents() with this async function
        # This is a limitation of the pyo3 async implementation
        async def upsert_documents():
            await self.collection.upsert_documents(documents, args)

        run_async_tasks([upsert_documents()])

    def add_documents(
        self,
        docs: Sequence[Document],
        **insert_kwargs: Any,
    ) -> None:
        nodes = [TextNode(**doc.dict()) for doc in docs]
        self._insert(nodes, **insert_kwargs)

    def delete_ref_doc(self, ref_doc_id: str) -> None:
        # We must wrap self.collection.delete_documents() with this async function
        # This is a limitation of the pyo3 async implementation
        async def delete_documents():
            await self.collection.delete_documents({"id": {"$eq": ref_doc_id}})

        run_async_tasks([delete_documents()])

    def update_ref_doc(self, document: Document) -> None:
        node = TextNode(**document.dict())
        self._insert([node], merge=True)

    def as_retriever(self, **kwargs: Any) -> BaseRetriever:
        """Return a Retriever for this managed index."""
        from llama_index.indices.managed.postgresml.retriever import (
            PostgresMLRetriever,
        )

        return PostgresMLRetriever(self, **kwargs)

    def as_query_engine(self, **kwargs: Any) -> BaseQueryEngine:
        from llama_index.indices.managed.postgresml.retriever import (
            PostgresMLRetriever,
        )
        from llama_index.indices.managed.postgresml.query import (
            PostgresMLQueryEngine,
        )

        return PostgresMLQueryEngine(PostgresMLRetriever(self, **kwargs), **kwargs)

    @classmethod
    def from_documents(
        cls: Type[IndexType],
        documents: Sequence[Document],
        collection_name: Optional[str] = None,
        pipeline_name: Optional[str] = None,
        pipeline_schema: Optional[Dict[str, Any]] = None,
        pgml_database_url: Optional[str] = None,
        show_progress: bool = False,
        upsert_parallel_batches: int = 1,
        **kwargs: Any,
    ) -> IndexType:
        """Build a PostgresML index from a sequence of documents."""
        if collection_name is None:
            raise Exception("collection_name is a required argument")
        nodes = [TextNode(**doc.dict()) for doc in documents]
        return cls(
            collection_name,
            pipeline_name=pipeline_name,
            pipeline_schema=pipeline_schema,
            pgml_database_url=pgml_database_url,
            nodes=nodes,
            show_progress=show_progress,
            upsert_parallel_batches=upsert_parallel_batches,
            **kwargs,
        )



from llama_index.indices.managed.postgresml.base import PostgresMLIndex
from llama_index.indices.managed.postgresml.retriever import PostgresMLRetriever

__all__ = ["PostgresMLIndex", "PostgresMLRetriever"]
